import pytest

from lasagna.lasagna_anthropic import (
    _convert_to_anthropic_messages,
    _build_messages_from_anthropic_payload,
    _convert_to_anthropic_tools,
    _make_raw_event,
    _process_stream,
)

from lasagna.types import (
    Message,
)

from lasagna.stream import fake_async

from typing import List

from anthropic.types.message import Message as AnthropicMessage
from anthropic.lib.streaming._types import (
    MessageStreamEvent,
    TextEvent,
    ContentBlockStopEvent,
)
from anthropic.types.raw_content_block_start_event import RawContentBlockStartEvent
from anthropic.types.raw_content_block_delta_event import RawContentBlockDeltaEvent

import os
import tempfile


@pytest.mark.asyncio
async def test_convert_to_anthropic_messages():
    # With system prompt:
    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'Hi, what are you?',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt == 'You are a robot.'
    assert anthropic_messages == [
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what are you?',
                },
            ],
        },
    ]

    # Without system prompt:
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
    ]

    # Media:
    with tempfile.TemporaryDirectory() as tmp:
        fn = os.path.join(tmp, 'a.png')
        with open(fn, 'wb') as f:
            f.write(b'1234')
        messages = [
            {
                'role': 'human',
                'text': "Hi, here's a picture.",
                'media': [
                    {
                        'type': 'image',
                        'image': fn,
                    },
                ],
            },
        ]
        system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
        assert system_prompt is None
        assert anthropic_messages == [
            {
                'role': 'user',
                'content': [
                    {
                        'type': 'text',
                        'text': "Hi, here's a picture.",
                    },
                    {
                        'type': 'image',
                        'source': {
                            'type': 'base64',
                            'media_type': 'image/png',
                            'data': 'MTIzNA==',
                        },
                    },
                ],
            },
        ]

    # Tool messages:
    messages: List[Message] = [
        {
            'role': 'tool_call',
            'tools': [
                {
                    'call_type': 'function',
                    'call_id': 'abcd',
                    'function': {
                        'name': 'get_stuff',
                        'arguments': '{"q": "stuff about cats"}',
                    },
                },
                {
                    'call_type': 'function',
                    'call_id': 'wxyz',
                    'function': {
                        'name': 'feed_cat',
                        'arguments': '{}',
                    },
                },
            ],
        },
        {
            'role': 'tool_res',
            'tools': [
                {
                    'call_id': 'abcd',
                    'result': 'here is your stuff',
                },
                {
                    'call_id': 'wxyz',
                    'result': 'Error: cat 404',
                    'is_error': True,
                },
            ],
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'tool_use',
                    'id': 'abcd',
                    'name': 'get_stuff',
                    'input': {'q': 'stuff about cats'},
                },
                {
                    'type': 'tool_use',
                    'id': 'wxyz',
                    'name': 'feed_cat',
                    'input': {},
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'type': 'tool_result',
                    'tool_use_id': 'abcd',
                    'content': [
                        {
                            'type': 'text',
                            'text': 'here is your stuff',
                        },
                    ],
                },
                {
                    'type': 'tool_result',
                    'tool_use_id': 'wxyz',
                    'content': [
                        {
                            'type': 'text',
                            'text': 'Error: cat 404',
                        },
                    ],
                    'is_error': True,
                },
            ],
        },
    ]

    # Invalid role:
    messages: List[Message] = [
        {
            'role': 'invalid',  # type: ignore
            'text': 'Hi, what can I help you with today?',
        },
    ]
    with pytest.raises(ValueError):
        await _convert_to_anthropic_messages(messages)

    # Collapse logic (most basic):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'ai',
            'text': 'Here is more text.',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what can I help you with today?',
                },
                {
                    'type': 'text',
                    'text': 'Here is more text.',
                },
            ],
        },
    ]

    # Collapse logic (ensure not to collapse in this case):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'I need help with everything.',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Please be more specific.',
                },
            ],
        },
    ]

    # Collapse logic (more complex):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'human',
            'text': 'EVERYTHING!',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
        {
            'role': 'ai',
            'text': 'Like tell me what you are trying to do.',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'I need help with everything.',
                },
                {
                    'type': 'text',
                    'text': 'EVERYTHING!',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Please be more specific.',
                },
                {
                    'type': 'text',
                    'text': 'Like tell me what you are trying to do.',
                },
            ],
        },
    ]

    # Collapse logic (more complex again):
    messages: List[Message] = [
        {
            'role': 'ai',
            'text': 'Hi, what can I help you with today?',
        },
        {
            'role': 'human',
            'text': 'I need help with everything.',
        },
        {
            'role': 'human',
            'text': 'EVERYTHING!',
        },
        {
            'role': 'ai',
            'text': 'Please be more specific.',
        },
        {
            'role': 'ai',
            'text': 'Like tell me what you are trying to do.',
        },
        {
            'role': 'human',
            'text': 'Forget it...',
        },
    ]
    system_prompt, anthropic_messages = await _convert_to_anthropic_messages(messages)
    assert system_prompt is None
    assert anthropic_messages == [
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Hi, what can I help you with today?',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'I need help with everything.',
                },
                {
                    'type': 'text',
                    'text': 'EVERYTHING!',
                },
            ],
        },
        {
            'role': 'assistant',
            'content': [
                {
                    'type': 'text',
                    'text': 'Please be more specific.',
                },
                {
                    'type': 'text',
                    'text': 'Like tell me what you are trying to do.',
                },
            ],
        },
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'Forget it...',
                },
            ],
        },
    ]


@pytest.mark.asyncio
async def test_build_messages_from_anthropic_payload():
    raw = {
        "id": "msg_01NvDCPEEZgGEzpvYnjqru9n",
        "content": [
            {
                "text": "Hello, I'm Bob. Let me help you with those calculations:",
                "type": "text",
            },
            {
                "id": "toolu_011nFpn7UokLDMfiP6UW7DuJ",
                "input": {
                    "a": 5,
                    "b": 81,
                },
                "name": "multiply",
                "type": "tool_use",
            },
            {
                "id": "toolu_01VhUoeyXmswsFtEc8piHi4b",
                "input": {
                    "a": 89,
                    "b": 412,
                },
                "name": "multiply",
                "type": "tool_use",
            },
        ],
        "model": "claude-3-haiku-20240307",
        "role": "assistant",
        "stop_reason": "tool_use",
        "stop_sequence": None,
        "type": "message",
        "usage": {
            "input_tokens": 508,
            "output_tokens": 82,
        },
    }
    anthropic_message = AnthropicMessage.model_validate(raw)
    messages = _build_messages_from_anthropic_payload([], anthropic_message)
    assert messages == [
        {
            'role': 'ai',
            'text': "Hello, I'm Bob. Let me help you with those calculations:",
        },
        {
            'role': 'tool_call',
            'tools': [
                {
                    'call_type': 'function',
                    'call_id': 'toolu_011nFpn7UokLDMfiP6UW7DuJ',
                    'function': {
                        'name': 'multiply',
                        'arguments': '{"a": 5, "b": 81}',
                    },
                },
                {
                    'call_type': 'function',
                    'call_id': 'toolu_01VhUoeyXmswsFtEc8piHi4b',
                    'function': {
                        'name': 'multiply',
                        'arguments': '{"a": 89, "b": 412}',
                    },
                },
            ],
            'cost': {
                'input_tokens': 508,
                'output_tokens': 82,
                'total_tokens': 590,
            },
            'raw': {
                'events': [],
                'message': raw,
            },
        },
    ]


def test_convert_to_anthropic_tools():
    def multiply(a, b):
        """
        Use this tool to get the product of two values.
        :param: a: float: the first value
        :param: b: float: the second value
        """
        return a * b
    assert _convert_to_anthropic_tools([multiply]) == [
        {
            'name': 'multiply',
            'description': 'Use this tool to get the product of two values.',
            'input_schema': {
                'type': 'object',
                'properties': {
                    'a': {'type': 'number', 'description': 'the first value'},
                    'b': {'type': 'number', 'description': 'the second value'},
                },
                'required': ['a', 'b'],
            },
        },
    ]


def test_make_raw_event():
    event = TextEvent.model_validate({
        'type': 'text',
        'text': '!',
        'snapshot': 'Hello!',
    })
    assert _make_raw_event(event) == {
        'type': 'text',
        'text': '!'
    }


@pytest.mark.asyncio
async def test_process_stream():
    events: List[MessageStreamEvent] = [
        RawContentBlockStartEvent.model_validate({
            "content_block": {
                "text": "",
                "type": "text"
            },
            "index": 0,
            "type": "content_block_start"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": "Hello",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": "!",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " I",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": "'m",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " Bob",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": ",",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " an",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " AI",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " assistant",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": ".",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " Let",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " me",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " help",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " you",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " with",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " those",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": " calculations",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "text": ":",
                "type": "text_delta"
            },
            "index": 0,
            "type": "content_block_delta"
        }),
        ContentBlockStopEvent.model_validate({
            "index": 0,
            "type": "content_block_stop",
            "content_block": {
                "text": "Hello! I'm Bob, an AI assistant. Let me help you with those calculations:",
                "type": "text"
            }
        }),
        RawContentBlockStartEvent.model_validate({
            "content_block": {
                "id": "toolu_01JxBsKfpT7qKZVhHnHub9tv",
                "input": {},
                "name": "multiply",
                "type": "tool_use"
            },
            "index": 1,
            "type": "content_block_start"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "partial_json": "",
                "type": "input_json_delta"
            },
            "index": 1,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "partial_json": "{\"a\": 5",
                "type": "input_json_delta"
            },
            "index": 1,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "partial_json": ", ",
                "type": "input_json_delta"
            },
            "index": 1,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "partial_json": "\"b\": 8",
                "type": "input_json_delta"
            },
            "index": 1,
            "type": "content_block_delta"
        }),
        RawContentBlockDeltaEvent.model_validate({
            "delta": {
                "partial_json": "1}",
                "type": "input_json_delta"
            },
            "index": 1,
            "type": "content_block_delta"
        }),
        ContentBlockStopEvent.model_validate({
            "index": 1,
            "type": "content_block_stop",
            "content_block": {
                "id": "toolu_01JxBsKfpT7qKZVhHnHub9tv",
                "input": {
                    "a": 5,
                    "b": 81
                },
                "name": "multiply",
                "type": "tool_use"
            }
        }),
    ]
    new_events = [e async for e in _process_stream(fake_async(events))]
    assert new_events == [
        ('ai', 'text_event', 'Hello'),
        ('ai', 'text_event', '!'),
        ('ai', 'text_event', ' I'),
        ('ai', 'text_event', "'m"),
        ('ai', 'text_event', ' Bob'),
        ('ai', 'text_event', ','),
        ('ai', 'text_event', ' an'),
        ('ai', 'text_event', ' AI'),
        ('ai', 'text_event', ' assistant'),
        ('ai', 'text_event', '.'),
        ('ai', 'text_event', ' Let'),
        ('ai', 'text_event', ' me'),
        ('ai', 'text_event', ' help'),
        ('ai', 'text_event', ' you'),
        ('ai', 'text_event', ' with'),
        ('ai', 'text_event', ' those'),
        ('ai', 'text_event', ' calculations'),
        ('ai', 'text_event', ':'),
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"a": 5'),
        ('tool_call', 'text_event', ', '),
        ('tool_call', 'text_event', '"b": 8'),
        ('tool_call', 'text_event', '1}'),
        ('tool_call', 'text_event', ')\n'),
        ('tool_call', 'tool_call_event', {'call_type': 'function', 'call_id': 'toolu_01JxBsKfpT7qKZVhHnHub9tv', 'function': {'name': 'multiply', 'arguments': '{"a": 5, "b": 81}'}}),
    ]



from lasagna import (
    bind_model,
    build_most_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable, cast

import asyncio

from dotenv import load_dotenv; load_dotenv()

import sympy as sp


MODEL_BINDER = bind_model('openai', 'gpt-3.5-turbo-0125')


def evaluate_math_expression(expression: str) -> float:
    """
    This tool evaluates a math expression and returns the result.
    Pass math expression as a string, for example:
     - "3 * 6 + 1"
     - "cos(2 * pi / 3) + log(8)"
     - "(4.5/2) + (6.3/1.2)"
     - ... etc
    :param: expression: str: the math expression to evaluate
    """
    expr = sp.sympify(expression)
    result = cast(float, expr.evalf())
    return result


async def main() -> None:
    system_prompt = "You are my math assistant."
    tools: List[Callable] = [
        evaluate_math_expression,
    ]
    my_agent = build_most_simple_agent(tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



from lasagna import (
    bind_model,
    build_most_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

import asyncio

from dotenv import load_dotenv; load_dotenv()


MODEL_BINDER = bind_model('openai', 'gpt-3.5-turbo-0125')


async def main() -> None:
    system_prompt = "You are grumpy."
    tools: List[Callable] = []
    my_agent = build_most_simple_agent(tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



import asyncio
import time
import functools

from typing import Callable, Union, List, Awaitable, Dict

from .types import (
    AgentCallable,
    AgentRun,
    Model,
    EventCallback,
    EventPayload,
    CacheKey,
    CacheRecord,
    CacheEventPayload,
)

from .util import recursive_hash

from . import __version__

import logging

_LOG = logging.getLogger(__name__)


async def _hash_agent_runs(model: Model, runs: List[AgentRun]) -> CacheKey:
    model_config_hash = model.config_hash()
    seed = f"__version__{__version__}__model__{model_config_hash}"
    def _do() -> CacheKey:
        return recursive_hash(seed, runs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(None, _do)


def cached_agent(
    query_record: Callable[[CacheKey], Awaitable[Union[CacheRecord, None]]],
    should_cache_record: Callable[[Model, List[AgentRun], AgentRun], Awaitable[bool]],
    save_record: Callable[[CacheKey, CacheRecord], Awaitable[None]],
    simulate_time: bool,
    hash_function: Callable[[Model, List[AgentRun]], Awaitable[CacheKey]] = _hash_agent_runs,
) -> Callable[[AgentCallable], AgentCallable]:
    def decorator(agent: AgentCallable) -> AgentCallable:
        @functools.wraps(agent, assigned=['__module__', '__name__', '__qualname__', '__doc__'])
        async def new_agent(model: Model, event_callback: EventCallback, prev_runs: List[AgentRun]) -> AgentRun:
            hash = await hash_function(model, prev_runs)
            old_cached_record = await query_record(hash)
            start_time = time.time()

            if old_cached_record is not None:
                _LOG.info(f"Cache **HIT**: {hash}")
                for event in old_cached_record['events']:
                    if simulate_time:
                        to_wait = event['delta_time'] - (time.time() - start_time)
                        if to_wait > 0.0:
                            await asyncio.sleep(to_wait)
                    await event_callback(event['event'])
                return old_cached_record['run']

            _LOG.info(f"Cache miss: {hash}")

            captured_events: List[CacheEventPayload] = []
            async def event_callback_wrapper(event: EventPayload) -> None:
                captured_events.append({
                    'delta_time': time.time() - start_time,
                    'event': event,
                })
                await event_callback(event)

            run = await agent(model, event_callback_wrapper, prev_runs)

            if await should_cache_record(model, prev_runs, run):
                _LOG.info(f"Will store in cache: {hash}")
                new_cached_record: CacheRecord = {
                    'events': captured_events,
                    'run': run,
                }
                await save_record(hash, new_cached_record)
            else:
                _LOG.info(f"Will *not* store in cache: {hash}")

            return run

        return new_agent

    return decorator


def in_memory_cached_agent(agent: AgentCallable) -> AgentCallable:
    """
    This is mostly just for demo to show how to use the @cached_agent
    decorator.
    A real system would use an external database, and would probably
    do some sort of LRU thing in the database so you don't cache too much.
    Also consider only caching *short* runs (because long runs are unlikely
    to every see a cache-hit, so it's not worth it).
    """
    cache: Dict[CacheKey, CacheRecord] = {}

    async def in_memory_query_record(key: CacheKey) -> Union[CacheRecord, None]:
        return cache.get(key)

    async def in_memory_should_cache_record(model: Model, prev_runs: List[AgentRun], this_run: AgentRun) -> bool:
        return len(cache) < 1000   # !!! just a demo, don't read into this

    async def in_memory_save_record(key: CacheKey, record: CacheRecord) -> None:
        cache[key] = record

    decorator = cached_agent(
        query_record = in_memory_query_record,
        should_cache_record = in_memory_should_cache_record,
        save_record = in_memory_save_record,
        simulate_time=True,
    )

    return decorator(agent)



from .registrar import register_provider

import logging

_LOG = logging.getLogger(__name__)


def attempt_load_all_known_providers() -> None:
    """
    Use this function when you want to populate the registrar
    at the start of the application. E.g. If you display a list of
    all known providers in the UI somewhere.
    """
    providers_to_try_to_load = [
        'openai',
        'anthropic',
        'nvidia',
    ]
    for provider in providers_to_try_to_load:
        try:
            attempt_load_known_providers(provider)
        except ImportError:
            _LOG.warning(f"failed to load provider: {provider}")


def attempt_load_known_providers(provider: str) -> None:
    """
    Attempt to load known providers at runtime, only when we *actually* need
    them.

    As the name implies, this is just an **attempt**.

    The `provider` string will be one of:
     1. known and loadable ✅, or...
     2. known but not loadable (probably due to a missing dependency) 😥, or...
     3. completely unknown 😥.

    Note: Normally I'd be sad by imports inside functions, but this is
          an acceptable use-case for doing so. We do *not* want to
          force people to install dependencies they don't need, thus
          we have to do some dynamic/runtime import-only-when-you-need-it
          behavior, which is what this function's job is. Moreover, that
          ugly behavior is restricted to *just this function* to keep the
          ugliness to a minimum.
    """
    if provider == 'openai':
        from .lasagna_openai import (
            OPENAI_KNOWN_MODELS,
            LasagnaOpenAI,
        )
        register_provider(
            key  = 'openai',
            name = 'OpenAI',
            factory = LasagnaOpenAI,
            models = OPENAI_KNOWN_MODELS,
        )

    elif provider == 'anthropic':
        from .lasagna_anthropic import (
            ANTHROPIC_KNOWN_MODELS,
            LasagnaAnthropic,
        )
        register_provider(
            key  = 'anthropic',
            name = 'Anthropic',
            factory = LasagnaAnthropic,
            models = ANTHROPIC_KNOWN_MODELS,
        )

    elif provider == 'nvidia':
        from .lasagna_nvidia import (
            NVIDIA_KNOWN_MODELS,
            LasagnaNVIDIA,
        )
        register_provider(
            key  = 'nvidia',
            name = 'NVIDIA',
            factory = LasagnaNVIDIA,
            models = NVIDIA_KNOWN_MODELS,
        )

    else:
        raise ValueError(f"unknown provider string: {provider}")



import functools

from typing import Union, Dict, Any, List, Callable

from .types import (
    AgentSpec,
    AgentRun,
    ModelFactory,
    AgentCallable,
    BoundAgentCallable,
    EventCallback,
    EventPayload,
    Message,
    Model,
)

from .agent_runner import run


def bind_model(
    provider: Union[str, ModelFactory],
    model: str,
    model_kwargs: Union[Dict[str, Any], None] = None,
) -> Callable[[AgentCallable], BoundAgentCallable]:
    def decorator(agent: AgentCallable) -> BoundAgentCallable:
        spec: AgentSpec = {
            'agent': agent,
            'provider': provider,
            'model': model,
            'model_kwargs': model_kwargs or {},
        }
        @functools.wraps(agent, assigned=['__module__', '__name__', '__qualname__', '__doc__'])
        async def bound_agent(event_callback: EventCallback, prev_runs: List[AgentRun]) -> AgentRun:
            return await run(spec, event_callback, prev_runs)
        return bound_agent
    return decorator


def recursive_extract_messages(agent_runs: List[AgentRun]) -> List[Message]:
    messages: List[Message] = []
    for run in agent_runs:
        if run['type'] == 'messages':
            messages.extend(run['messages'])
        else:
            messages.extend(recursive_extract_messages(run['runs']))
    return messages


def flat_messages(messages: List[Message]) -> AgentRun:
    return {
        'type': 'messages',
        'messages': messages,
    }


def build_most_simple_agent(
    tools: List[Callable] = [],
) -> AgentCallable:
    async def most_simple_agent(
        model: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRun:
        messages = recursive_extract_messages(prev_runs)
        new_messages = await model.run(event_callback, messages, tools)
        return flat_messages(new_messages)

    return most_simple_agent


async def noop_callback(event: EventPayload) -> None:
    # "noop" mean "no operation" means DON'T DO ANYTHING!
    pass


def extract_last_message(
    agent_run_or_runs: Union[AgentRun, List[AgentRun]],
) -> Message:
    if isinstance(agent_run_or_runs, list):
        messages = recursive_extract_messages(agent_run_or_runs)
    else:
        messages = recursive_extract_messages([agent_run_or_runs])
    if len(messages) == 0:
        raise ValueError('no messages found')
    return messages[-1]



import asyncio
import inspect
import json

from typing import List, Dict, Callable, Union, Any

from .types import (
    Message,
    ToolCall,
    ToolResult,
    ToolParam,
)


def convert_to_json_schema(params: List[ToolParam]) -> Dict[str, object]:
    def convert_type(t: str) -> Dict[str, object]:
        if t.startswith('enum '):
            return {
                "type": "string",
                "enum": t.split()[1:],
            }
        else:
            return {
                "type": {
                    'str': 'string',
                    'float': 'number',
                    'int': 'integer',
                    'bool': 'boolean',
                }[t],
            }
    return {
        "type": "object",
        "properties": {
            p['name']: {
                **convert_type(p['type']),
                "description": p['description'],
            }
            for p in params
        },
        "required": [
            p['name']
            for p in params
            if not p['description'].startswith('(optional)')
        ],
    }


async def handle_tools(
    messages: List[Message],
    tools_map: Dict[str, Callable],
) -> Union[List[ToolResult], None]:
    assert len(messages) > 0
    tool_messages = [message for message in messages if message['role'] == 'tool_call']
    if len(tool_messages) == 0:
        return None
    to_gather: List[asyncio.Task[ToolResult]] = []
    for message in tool_messages:
        for t in message['tools']:
            assert t['call_type'] == 'function'
            async def _go(t: ToolCall) -> ToolResult:
                call_id = 'unknown'
                try:
                    call_id = t['call_id']
                    func = tools_map[t['function']['name']]
                    args = t['function']['arguments']
                    if inspect.iscoroutinefunction(func):
                        res = await func(**json.loads(args))
                    else:
                        def _wrapped_sync() -> Any:
                            return func(**json.loads(args))
                        loop = asyncio.get_running_loop()
                        res = await loop.run_in_executor(None, _wrapped_sync)
                    return {'call_id': call_id, 'result': res}
                except Exception as e:
                    error = f"{type(e).__name__}: {e}"
                    return {'call_id': call_id, 'result': error, 'is_error': True}
            to_gather.append(asyncio.create_task(_go(t)))
    return await asyncio.gather(*to_gather)


def build_tool_response_message(tool_results: List[ToolResult]) -> Message:
    return {
        'role': 'tool_res',
        'tools': tool_results,
    }



import pytest
import hashlib
import random
from typing import List

from lasagna.caching import (
    in_memory_cached_agent,
    _hash_agent_runs,
)

from lasagna.types import AgentRun, Model, EventCallback, Message

from lasagna.agent_util import bind_model

from lasagna import __version__

from lasagna.mock_provider import (
    MockProvider,
)


_PREV_RUNS: List[AgentRun] = [
    {
        'type': 'messages',
        'messages': [
            {
                'role': 'system',
                'text': 'You are a robot.',
            },
        ],
    },
]


def _make_agent():
    # We need this factory so we can make as many separately-cached agents as we want.
    @bind_model(MockProvider, 'some_model')
    @in_memory_cached_agent
    async def my_agent(
        model: Model,
        event_callback: EventCallback,
        prev_runs: List[AgentRun],
    ) -> AgentRun:
        assert prev_runs == _PREV_RUNS
        messages: List[Message] = []
        new_messages = await model.run(event_callback, messages, [])
        return {
            'type': 'messages',
            'messages': [
                *new_messages,
                {
                    'role': 'human',
                    'text': f"Here is a random value {random.random()}.",
                },
            ],
        }

    return my_agent


@pytest.mark.asyncio
async def test_in_memory_cached_agent():
    """
    We test the *in-memory* cache just as a concrete way to test the
    underlying cache decorator.
    """
    async def callback(event):
        pass

    agent1 = _make_agent()
    result1 = await agent1(callback, _PREV_RUNS)

    agent2 = _make_agent()
    result2 = await agent2(callback, _PREV_RUNS)

    assert result1 != result2   # because of the use of random.random() in the final message, these will be different (in the *vast* likelihood)

    for _ in range(5):
        result1_again = await agent1(callback, _PREV_RUNS)
        result2_again = await agent2(callback, _PREV_RUNS)

        assert result1 == result1_again  # because it was cached
        assert result2 == result2_again  # because it was cached


@pytest.mark.asyncio
async def test_hash_agent_runs():
    model = MockProvider(model='something')

    s = ''.join([v.strip() for v in f'''
        __open_dict__
            __open_list__
                __open_list__
                    __str__model
                    __str__something
                __close_list__
                __open_list__
                    __str__model_kwargs
                    __open_dict__
                        __open_list__
                        __close_list__
                    __close_dict__
                __close_list__
                __open_list__
                    __str__provider
                    __str__mock
                __close_list__
            __close_list__
        __close_dict__
        '''.strip().splitlines()])

    model_config_hash = hashlib.md5(s.encode('utf-8')).hexdigest()

    got = await _hash_agent_runs(model, _PREV_RUNS)

    s = ''.join([v.strip() for v in f'''
        __str____version__{__version__}__model__{model_config_hash}
        __open_list__
            __open_dict__
                __open_list__
                    __open_list__
                        __str__messages
                        __open_list__
                            __open_dict__
                                __open_list__
                                    __open_list__
                                        __str__role
                                        __str__system
                                    __close_list__
                                    __open_list__
                                        __str__text
                                        __str__You are a robot.
                                    __close_list__
                                __close_list__
                            __close_dict__
                        __close_list__
                    __close_list__
                    __open_list__
                        __str__type
                        __str__messages
                    __close_list__
                __close_list__
            __close_dict__
        __close_list__
        '''.strip().splitlines()])

    correct = hashlib.md5(s.encode('utf-8')).hexdigest()

    assert got == correct



from lasagna.types import (
    EventCallback,
    Message,
    Model,
    EventPayload,
)

from lasagna.util import recursive_hash

from typing import List, Dict, Any, Callable


class MockProvider(Model):
    def __init__(self, model: str, **model_kwargs: Dict[str, Any]):
        self.model = model
        self.model_kwargs = model_kwargs

    def config_hash(self) -> str:
        return recursive_hash(None, {
            'provider': 'mock',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        })

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        event: EventPayload = 'ai', 'text_event', 'Hi!'
        await event_callback(event)
        res: List[Message] = [
            {
                'role': 'ai',
                'text': f"model: {self.model}",
                'cost': None,
                'raw': None,
            },
        ]
        for key in sorted(self.model_kwargs.keys()):
            val = self.model_kwargs[key]
            m: Message = {
                'role': 'human',
                'text': f"model_kwarg: {key} = {val}",
                'cost': None,
                'raw': None,
            }
            res.append(m)
        return res



from lasagna import (
    bind_model,
    flat_messages,
    build_most_simple_agent,
    extract_last_message,
    noop_callback,
)

from lasagna.types import Message

from typing import List, Dict

import asyncio

from dotenv import load_dotenv; load_dotenv()


THE_AGENT = build_most_simple_agent()


COMMITTEE_SPEC = [
    ('openai',    'gpt-4o-2024-05-13'),
    ('openai',    'gpt-4-turbo-2024-04-09'),
    ('openai',    'gpt-3.5-turbo-0125'),
    ('anthropic', 'claude-3-5-sonnet-20240620'),
    ('anthropic', 'claude-3-opus-20240229'),
    ('anthropic', 'claude-3-sonnet-20240229'),
    ('anthropic', 'claude-3-haiku-20240307'),
    ('nvidia',    'meta/llama3-70b-instruct'),
    ('nvidia',    'meta/llama3-8b-instruct'),
    ('nvidia',    'mistralai/mistral-large'),
    ('nvidia',    'mistralai/mixtral-8x7b-instruct-v0.1'),
    ('nvidia',    'google/recurrentgemma-2b'),
    ('nvidia',    'microsoft/phi-3-mini-128k-instruct'),
    ('nvidia',    'snowflake/arctic'),
]


COMMITTEE_MODELS = [
    bind_model(*model)(THE_AGENT)
    for model in COMMITTEE_SPEC
]


async def vote_on_jokes(joke_a: str, joke_b: str) -> Dict[str, int]:
    messages: List[Message] = [
        {
            'role': 'system',
            'text': "The user will provider you with two jokes named `a` and `b`. You will output which is funniest. Output just a single character (`a` or `b`) to cast your vote.",
        },
        {
            'role': 'human',
            'text': f"a: {joke_a}\n\nb: {joke_b}\n\nRemember, output **only** a single token a or b without any preamble.",
        },
    ]
    tasks = [
        model(noop_callback, [flat_messages(messages)])
        for model in COMMITTEE_MODELS
    ]
    outputs = await asyncio.gather(*tasks)
    counter: Dict[str, int] = {}
    for spec, out in zip(COMMITTEE_SPEC, outputs):
        last_message = extract_last_message(out)
        assert last_message['role'] == 'ai'
        text = last_message['text']
        assert text
        text = text.strip()
        print(spec, text)
        if text not in counter:
            counter[text] = 0
        counter[text] += 1
    print('---------')
    return counter


async def main() -> None:
    #joke_a = "Why don't skeletons fight each other? They don't have the guts."
    #joke_a = "Why did the chicken cross the road? To get to the other side!"
    #joke_a = "I'm not funny."
    joke_a = "Why did the cat jump? Because cats like to jump..."
    joke_b = "Why do programmers prefer dark mode? Because light attracts bugs!"
    votes1 = await vote_on_jokes(joke_a, joke_b)
    votes2 = await vote_on_jokes(joke_b, joke_a)
    votes = {
        'a': (votes1.get('a', 0) + votes2.get('b', 0)),
        'b': (votes1.get('b', 0) + votes2.get('a', 0)),
    }
    print(votes)


if __name__ == '__main__':
    asyncio.run(main())



import pytest

from lasagna.types import (
    Message,
    ToolResult,
)

from typing import List, Dict, Callable

from lasagna.tools_util import (
    convert_to_json_schema,
    handle_tools,
    build_tool_response_message,
)


def test_convert_to_json_schema():
    j = convert_to_json_schema([
        {
            'name': 'a',
            'type': 'str',
            'description': 'param a',
        },
        {
            'name': 'second',
            'type': 'float',
            'description': 'second param',
        },
        {
            'name': 'another',
            'type': 'int',
            'description': 'another param',
        },
        {
            'name': 'yet_another',
            'type': 'bool',
            'description': 'yet another param',
        },
        {
            'name': 'an_enum',
            'type': 'enum cat dog house',
            'description': 'an enum param',
        },
        {
            'name': 'last_param',
            'type': 'bool',
            'description': '(optional) you can pass this if you want',
        },
    ])
    assert j == {
        'type': 'object',
        'properties': {
            'a': {
                'type': 'string',
                'description': 'param a',
            },
            'second': {
                'type': 'number',
                'description': 'second param',
            },
            'another': {
                'type': 'integer',
                'description': 'another param',
            },
            'yet_another': {
                'type': 'boolean',
                'description': 'yet another param',
            },
            'an_enum': {
                'type': 'string',
                'enum': ['cat', 'dog', 'house'],
                'description': 'an enum param',
            },
            'last_param': {
                'type': 'boolean',
                'description': '(optional) you can pass this if you want',
            },
        },
        'required': ['a', 'second', 'another', 'yet_another', 'an_enum'],
    }


def tool_a(first, second, third=5):
    return first + second + third

def tool_b(x):
    return x * 2

async def tool_async_a(x):
    return x * 3


@pytest.mark.asyncio
async def test_handle_tools():
    x = 4
    def tool_c():
        return x * 4

    async def tool_async_b():
        return x * 5

    tool_map: Dict[str, Callable] = {
        'tool_a': tool_a,
        'tool_b': tool_b,
        'tool_c': tool_c,
        'tool_async_a': tool_async_a,
        'tool_async_b': tool_async_b,
    }
    message: Message = {
        'role': 'tool_call',
        'tools': [
            {'call_id': '1001', 'function': {'arguments': '{"x": 8}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1002', 'function': {'arguments': '{"x": 5.4}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1003', 'function': {'arguments': '{"x": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1004', 'function': {'arguments': '{}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1005', 'function': {'arguments': '{"y": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1006', 'function': {'arguments': '{"x": 4, "y": "hi"}', 'name': 'tool_b'}, 'call_type': 'function'},
            {'call_id': '1007', 'function': {'arguments': '{"first": 5, "second": 7.5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1008', 'function': {'arguments': '{"first": 5, "second": 7.5, "third": 3}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1009', 'function': {'arguments': '{"third": 99, "first": 5, "second": 7.5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1010', 'function': {'arguments': '{}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1011', 'function': {'arguments': '{"first": 5}', 'name': 'tool_a'}, 'call_type': 'function'},
            {'call_id': '1012', 'function': {'arguments': '{}', 'name': 'tool_c'}, 'call_type': 'function'},
            {'call_id': '1013', 'function': {'arguments': '{}', 'name': 'tool_d'}, 'call_type': 'function'},
            {'call_id': '1014', 'function': {'arguments': '{"x": -3}', 'name': 'tool_async_a'}, 'call_type': 'function'},
            {'call_id': '1015', 'function': {'arguments': '{}', 'name': 'tool_async_b'}, 'call_type': 'function'},
            {'call_id': '1016', 'function': {'arguments': '{}', 'name': 'tool_async_a'}, 'call_type': 'function'},
        ],
        'cost': None,
        'raw': None,
    }
    tool_results = await handle_tools([message], tool_map)
    assert tool_results is not None
    assert tool_results == [
        {'call_id': '1001', 'result': 16 },
        {'call_id': '1002', 'result': 10.8 },
        {'call_id': '1003', 'result': "hihi" },
        {'call_id': '1004', 'result': "TypeError: tool_b() missing 1 required positional argument: 'x'", 'is_error': True },
        {'call_id': '1005', 'result': "TypeError: tool_b() got an unexpected keyword argument 'y'", 'is_error': True },
        {'call_id': '1006', 'result': "TypeError: tool_b() got an unexpected keyword argument 'y'", 'is_error': True },
        {'call_id': '1007', 'result': 17.5 },
        {'call_id': '1008', 'result': 15.5 },
        {'call_id': '1009', 'result': 111.5 },
        {'call_id': '1010', 'result': "TypeError: tool_a() missing 2 required positional arguments: 'first' and 'second'", 'is_error': True },
        {'call_id': '1011', 'result': "TypeError: tool_a() missing 1 required positional argument: 'second'", 'is_error': True },
        {'call_id': '1012', 'result': 16 },
        {'call_id': '1013', 'result': "KeyError: 'tool_d'", 'is_error': True },
        {'call_id': '1014', 'result': -9 },
        {'call_id': '1015', 'result': 20 },
        {'call_id': '1016', 'result': "TypeError: tool_async_a() missing 1 required positional argument: 'x'", 'is_error': True },
    ]


def test_build_tool_response_message():
    res: List[ToolResult] = [
        {'call_id': '1002', 'result': 10.8 },
        {'call_id': '1003', 'result': "hihi" },
    ]
    message = build_tool_response_message(res)
    assert message == {
        'role': 'tool_res',
        'tools': res,
    }



from .agent_util import flat_messages

from .types import (
    EventPayload,
    BoundAgentCallable,
    AgentRun,
)

from typing import Union, List

from colorama import just_fix_windows_console, Fore, Style


async def tui_event_callback(event: EventPayload) -> None:
    just_fix_windows_console()
    if event[0] == 'ai' and event[1] == 'text_event':
        t = Style.RESET_ALL + event[2]
        print(t, end='', flush=True)
    elif event[0] == 'tool_call' and event[1] == 'text_event':
        assert isinstance(event[2], str)  # <-- mypy should *know* this, but it doesn't for some reason
        s = Fore.RED + event[2]
        print(s, end='', flush=True)
    elif event[0] == 'tool_res' and event[1] == 'tool_res_event':
        r = Fore.BLUE + f" -> {event[2]['result']}"
        print(r)


async def tui_input_loop(
    bound_agent: BoundAgentCallable,
    system_prompt: Union[str, None] = None,
) -> None:
    just_fix_windows_console()
    prev_runs: List[AgentRun] = []
    if system_prompt is not None:
        prev_runs.append(flat_messages([{
            'role': 'system',
            'text': system_prompt,
        }]))
    try:
        while True:
            human_input = input(Fore.GREEN + Style.BRIGHT + '> ')
            print(Style.RESET_ALL, end='', flush=True)
            prev_runs.append(flat_messages([{
                'role': 'human',
                'text': human_input,
            }]))
            this_run = await bound_agent(tui_event_callback, prev_runs)
            prev_runs.append(this_run)
            print(Style.RESET_ALL)
    except EOFError:
        # User hit ctrl-d
        pass
    except KeyboardInterrupt:
        # User hit ctrl-c
        pass
    finally:
        print(Style.RESET_ALL)



from lasagna import (
    bind_model,
    build_most_simple_agent,
    flat_messages,
)

from lasagna.tui import (
    tui_event_callback,
)

from lasagna.types import Message

from typing import List

from dotenv import load_dotenv; load_dotenv()

import asyncio


async def main() -> None:
    messages: List[Message] = [
        {
            'role': 'human',
            'text': "Describe this image:",
            'media': [
                {
                    'type': 'image',
                    'image': 'https://raw.githubusercontent.com/Rhobota/lasagna-ai/main/logos/lasagna-ai.png',
                },
            ],
        },
    ]
    bound_agent = bind_model('anthropic', 'claude-3-5-sonnet-20240620')(
        build_most_simple_agent(),
    )
    await bound_agent(tui_event_callback, [flat_messages(messages)])
    print()


if __name__ == '__main__':
    asyncio.run(main())



from lasagna.known_providers import attempt_load_all_known_providers

from lasagna.registrar import (
    AGENTS,
    PROVIDERS,
)


def test_attempt_load_known_providers():
    AGENTS.clear()
    PROVIDERS.clear()
    attempt_load_all_known_providers()
    assert len(PROVIDERS) == 3
    assert len(AGENTS) == 0



from .types import *

from .agent_runner import run

from .agent_util import (
    bind_model,
    recursive_extract_messages,
    flat_messages,
    build_most_simple_agent,
    noop_callback,
    extract_last_message,
)

__all__ = [
    'run',
    'bind_model',
    'recursive_extract_messages',
    'flat_messages',
    'build_most_simple_agent',
    'noop_callback',
    'extract_last_message',
]

__version__ = "0.7.1"



import pytest

from typing import List

from lasagna.agent_util import (
    bind_model,
    recursive_extract_messages,
    flat_messages,
)

from lasagna.types import (
    Model,
    EventCallback,
    AgentRun,
    Message,
    EventPayload,
)

from lasagna.mock_provider import (
    MockProvider,
)


@bind_model(MockProvider, 'some_model', {'a': 'yes', 'b': 6})
async def my_agent(
    model: Model,
    event_callback: EventCallback,
    prev_runs: List[AgentRun],
) -> AgentRun:
    assert len(prev_runs) == 0
    messages: List[Message] = []
    new_messages = await model.run(event_callback, messages, [])
    return {
        'type': 'messages',
        'messages': new_messages,
    }


@pytest.mark.asyncio
async def test_bind_model():
    events = []
    async def event_callback(event: EventPayload) -> None:
        events.append(event)
    prev_runs: List[AgentRun] = []
    new_run = await my_agent(event_callback, prev_runs)
    assert new_run == {
        'agent': 'my_agent',
        'provider': 'MockProvider',
        'model': 'some_model',
        'model_kwargs': {
            'b': 6,
            'a': 'yes',
        },
        'type': 'messages',
        'messages': [
            {
                'role': 'ai',
                'text': f"model: some_model",
                'cost': None,
                'raw': None,
            },
            {
                'role': 'human',
                'text': f"model_kwarg: a = yes",
                'cost': None,
                'raw': None,
            },
            {
                'role': 'human',
                'text': f"model_kwarg: b = 6",
                'cost': None,
                'raw': None,
            },
        ],
    }
    assert events == [
        ('ai', 'text_event', 'Hi!'),
    ]


def test_recursive_extract_messages():
    agent_run: AgentRun = {
        'type': 'chain',
        'runs': [
            {
                'type': 'parallel',
                'runs': [
                    {
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'system',
                                'text': 'You are a robot.',
                            },
                            {
                                'role': 'human',
                                'text': 'What are you?',
                            },
                        ],
                    },
                    {
                        'type': 'messages',
                        'messages': [
                            {
                                'role': 'system',
                                'text': 'You are a cat.',
                            },
                            {
                                'role': 'human',
                                'text': 'Here kitty kitty!',
                            },
                        ],
                    },
                ],
            },
            {
                'type': 'messages',
                'messages': [
                    {
                        'role': 'system',
                        'text': 'You aggregate other AI systems.',
                    },
                    {
                        'role': 'human',
                        'text': 'Summarize the previous AI conversations, please.',
                    },
                ],
            },
        ],
    }
    assert recursive_extract_messages([agent_run]) == [
        {
            'role': 'system',
            'text': 'You are a robot.',
        },
        {
            'role': 'human',
            'text': 'What are you?',
        },
        {
            'role': 'system',
            'text': 'You are a cat.',
        },
        {
            'role': 'human',
            'text': 'Here kitty kitty!',
        },
        {
            'role': 'system',
            'text': 'You aggregate other AI systems.',
        },
        {
            'role': 'human',
            'text': 'Summarize the previous AI conversations, please.',
        },
    ]


def test_flat_messages():
    messages: List[Message] = [
        {
            'role': 'system',
            'text': 'You are a cat.',
        },
        {
            'role': 'human',
            'text': 'Here kitty kitty!',
        },
    ]
    assert flat_messages(messages) == {
        'type': 'messages',
        'messages': [
            {
                'role': 'system',
                'text': 'You are a cat.',
            },
            {
                'role': 'human',
                'text': 'Here kitty kitty!',
            },
        ],
    }



from lasagna import (
    bind_model,
    build_most_simple_agent,
)

from lasagna.tui import (
    tui_input_loop,
)

from typing import List, Callable

from dotenv import load_dotenv; load_dotenv()

from twilio.rest import Client as TwilioClient   # type: ignore

import asyncio
import os


MODEL_BINDER = bind_model('openai', 'gpt-4o-2024-05-13')


def send_sms(to_phone_number: str, message_body: str) -> None:
    """
    Use this tool to send SMS message (aka, "text messages").
    :param: to_phone_number: str: the destination phone number to send to (in E.164 format; e.g. '+12223334444', where '+1' is the country code, '222' is the area code and '3334444' is the 7-digit phone number)
    :param: message_body: str: the content (aka, "body") of the SMS message
    """
    client = TwilioClient(
        os.environ['TWILIO_ACCOUNT_SID'],
        os.environ['TWILIO_AUTH_TOKEN'],
    )
    client.messages.create(
        body=message_body,
        from_=os.environ['TWILIO_FROM_PHONE'],
        to=to_phone_number,
    )


async def main() -> None:
    system_prompt = "You are grumpy."
    tools: List[Callable] = [
        send_sms,
    ]
    my_agent = build_most_simple_agent(tools)
    my_bound_agent = MODEL_BINDER(my_agent)
    await tui_input_loop(my_bound_agent, system_prompt)


if __name__ == '__main__':
    asyncio.run(main())



"""
This module is the Lasagna adapter for the NVIDIA NIM/NGC service.

For more information about the NVIDIA services this adapter is for, see:
 - https://org.ngc.nvidia.com/
 - https://build.nvidia.com/explore/reasoning
 - https://docs.nvidia.com/nim/large-language-models/latest/getting-started.html
"""

from .types import (
    Message,
    MessageContent,
    MessageToolCall,
    EventCallback,
    EventPayload,
    Model,
    ToolCall,
    ModelRecord,
    Cost,
)

from .stream import (
    apeek,
    adup,
    prefix_stream,
)

from .util import (
    parse_docstring,
    combine_pairs,
    convert_to_image_url,
    exponential_backoff_retry_delays,
    recursive_hash,
)

from .tools_util import (
    convert_to_json_schema,
    handle_tools,
    build_tool_response_message,
)

from openai import AsyncOpenAI, NOT_GIVEN, NotGiven
from openai.types.chat import (
    ChatCompletionChunk,
    ChatCompletionToolParam,
    ChatCompletionToolChoiceOptionParam,
    ChatCompletionMessageParam,
    ChatCompletionContentPartParam,
)
from openai.types.chat.chat_completion_chunk import ChoiceDelta
from openai import APIError

from typing import (
    List, Callable, AsyncIterator, Any, cast,
    Tuple, Dict, Optional, Union, Literal,
)

import asyncio
import copy
import json
import os

import logging

_LOG = logging.getLogger(__name__)


NVIDIA_KNOWN_MODELS: List[ModelRecord] = [
    {
        'formal_name': 'meta/llama3-70b-instruct',
        'display_name': 'meta/llama3-70b-instruct',
    },
    {
        'formal_name': 'meta/llama3-8b-instruct',
        'display_name': 'meta/llama3-8b-instruct',
    },
    {
        'formal_name': 'mistralai/mistral-large',
        'display_name': 'mistralai/mistral-large',
    },
    {
        'formal_name': 'mistralai/codestral-22b-instruct-v0.1',
        'display_name': 'mistralai/codestral-22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x22b-instruct-v0.1',
    },
    {
        'formal_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
        'display_name': 'mistralai/mixtral-8x7b-instruct-v0.1',
    },
    {
        'formal_name': 'google/gemma-7b',
        'display_name': 'google/gemma-7b',
    },
    {
        'formal_name': 'google/recurrentgemma-2b',
        'display_name': 'google/recurrentgemma-2b',
    },
    {
        'formal_name': 'microsoft/phi-3-mini-128k-instruct',
        'display_name': 'microsoft/phi-3-mini-128k-instruct',
    },
    {
        'formal_name': 'snowflake/arctic',
        'display_name': 'snowflake/arctic',
    },
]


async def _process_text_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    async for delta, finish_reason in stream:
        text = delta.content
        if finish_reason is not None:
            if text is not None:
                yield 'ai', 'text_event', str(text)
            return
        elif delta.tool_calls is not None:
            # The model is switching from text to tools!
            yield 'ai', 'text_event', "\n\n"
            put_back_val: Tuple[ChoiceDelta, Union[str, None]] = (delta, finish_reason)
            fixed_stream = prefix_stream([put_back_val], stream)
            substream = _process_tool_call_stream(fixed_stream)
            async for subval in substream:
                yield subval
            return
        elif text is not None:
            yield 'ai', 'text_event', str(text)


async def _process_tool_call_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    recs_by_index: Dict[int, ToolCall] = {}
    args_by_index: Dict[int, List[str]] = {}
    last_index: Optional[int] = None
    async for delta, _ in stream:
        if not delta.tool_calls:
            continue
        for tc in delta.tool_calls:
            index = tc.index
            if index != last_index and last_index is not None:
                yield 'tool_call', 'text_event', ")\n"   # <-- again, assumes no index-interleave
            last_index = index
            if index not in recs_by_index:
                assert tc.type == 'function', f"The only tool type we can do is a function! But got: {tc.type}"
                assert tc.function
                assert tc.function.name
                assert tc.id
                n: str = tc.function.name
                a: str = tc.function.arguments or ''
                recs_by_index[index] = {
                    'call_id': tc.id,
                    'call_type': 'function',
                    'function': {
                        'name': n,
                        'arguments': '',  # will be filled in at the end
                    },
                }
                args_by_index[index] = [a]
                yield 'tool_call', 'text_event', f"{n}("   # <-- assumes no index-interleave
            else:
                # assumes nothing but the argument is in the delta message...
                args = args_by_index[index]
                assert tc.function
                assert tc.function.arguments
                assert not tc.function.name
                a_delta: str = tc.function.arguments
                args.append(a_delta)
                yield 'tool_call', 'text_event', a_delta
    if last_index is not None:
        yield 'tool_call', 'text_event', ")"   # <-- again, assumes no index-interleave
    for index in sorted(recs_by_index.keys()):
        rec = recs_by_index[index]
        rec['function']['arguments'] = ''.join(args_by_index[index])
        yield 'tool_call', 'tool_call_event', rec


async def _process_output_stream(
    stream: AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]],
) -> AsyncIterator[EventPayload]:
    first, stream = await apeek(stream, n=1)
    first_delta, _ = first[0]
    is_text = first_delta.tool_calls is None   # <-- hacky, but works?
    if is_text:
        gen = _process_text_stream
    else:
        gen = _process_tool_call_stream
    async for v in gen(stream):
        yield v


async def _extract_deltas(
    stream: AsyncIterator[ChatCompletionChunk],
) -> AsyncIterator[Tuple[ChoiceDelta, Union[str, None]]]:
    async for v in stream:
        if len(v.choices) == 0:
            # The final message that has the `usage` has zero choices.
            # So just skip it here!
            continue
        assert len(v.choices) == 1, f"Why do we have {len(v.choices)} choices?"
        single_choice = v.choices[0]
        yield single_choice.delta, single_choice.finish_reason


def _convert_to_openai_tool(tool: Callable) -> ChatCompletionToolParam:
    description, params = parse_docstring(tool.__doc__ or '')
    return {
        'type': 'function',
        'function': {
            'name': tool.__name__,
            'description': description,
            'parameters': convert_to_json_schema(params),
        },
    }


def _convert_to_openai_tools(tools: List[Callable]) -> Union[NotGiven, List[ChatCompletionToolParam]]:
    if len(tools) == 0:
        return NOT_GIVEN
    specs = [_convert_to_openai_tool(tool) for tool in tools]
    return specs


async def _make_openai_content(
    message: MessageContent,
) -> Union[str, List[ChatCompletionContentPartParam]]:
    ret: List[ChatCompletionContentPartParam] = []
    if message['text'] and not message.get('media'):
        return message['text']
    elif message['text']:
        ret.append({
            'type': 'text',
            'text': message['text'],
        })
    if 'media' in message:
        for m in message['media']:
            if m['type'] == 'image':
                ret.append({
                    'type': 'image_url',
                    'image_url': {
                        'url': (await convert_to_image_url(m['image'])),
                    },
                })
            else:
                raise ValueError(f"unknown media type: {m['type']}")
    if len(ret) == 0:
        raise ValueError('no content in this message!')
    return ret


async def _convert_to_openai_messages(messages: List[Message]) -> List[ChatCompletionMessageParam]:
    ms: List[ChatCompletionMessageParam] = []
    for m in messages:
        if m['role'] == 'tool_call':
            tool_calls = m['tools']
            for tool_call in tool_calls:
                assert tool_call['call_type'] == 'function'
            ms.append({
                'role': 'assistant',
                'content': None,
                'tool_calls': [
                    {
                        'id': tool_call['call_id'],
                        'type': tool_call['call_type'],
                        'function': {
                            'name': tool_call['function']['name'],
                            'arguments': tool_call['function']['arguments'],
                        },
                    }
                    for tool_call in tool_calls
                ],
            })
        elif m['role'] == 'tool_res':
            tool_results = m['tools']
            for tool_result in tool_results:
                ms.append({
                    'role': 'tool',
                    'content': str(tool_result['result']),
                    'tool_call_id': tool_result['call_id'],
                })
        elif m['role'] == 'system':
            if 'media' in m and len(m['media']) > 0:
                raise ValueError('This model does not support media in the system prompt.')
            ms.append({
                'role': 'system',
                'content': m['text'] or '',
            })
        elif m['role'] == 'human':
            ms.append({
                'role': 'user',
                'content': (await _make_openai_content(m)),
            })
        elif m['role'] == 'ai':
            if 'media' in m and len(m['media']) > 0:
                raise ValueError('This model does not support media in AI messages.')
            ms.append({
                'role': 'assistant',
                'content': m['text'],
            })
        else:
            raise ValueError(f"unknown message role: {m['role']}")
    def should_combine(
        m1: ChatCompletionMessageParam,
        m2: ChatCompletionMessageParam,
    ) -> Union[Literal[False], Tuple[Literal[True], ChatCompletionMessageParam]]:
        if m1['role'] == 'assistant' and m2['role'] == 'assistant':
            # This is the case where the model started with text and switched
            # to tool-calling part-way-through. We need to combine these
            # messages.
            assert ('content' in m1 and m1['content']) and ('tool_calls' not in m1 or not m1['tool_calls'])
            assert ('content' not in m2 or not m2['content']) and ('tool_calls' in m2 and m2['tool_calls'])
            m_combined: ChatCompletionMessageParam = {
                'role': 'assistant',
                'content': m1['content'],
                'tool_calls': m2['tool_calls'],
            }
            return True, m_combined
        return False
    return combine_pairs(ms, should_combine)


def _get_cost(
    payload: List[ChatCompletionChunk],
) -> Optional[Cost]:
    usages = [p.usage for p in payload if p.usage]
    if not usages:
        return None
    usage = usages[-1]
    return {
        'input_tokens': usage.prompt_tokens,
        'output_tokens': usage.completion_tokens,
        'total_tokens': usage.total_tokens,
    }


def _build_messages_from_openai_payload(
    payload: List[ChatCompletionChunk],
    events: List[EventPayload],
) -> List[Message]:
    """
    We either have:
     - all AI events
     - all TOOL_CALL events
     - some AI events then it switches to TOOL_CALL events
    """
    cost = _get_cost(payload)
    raw = [p.to_dict() for p in payload]
    ai_events = [
        event
        for event in events
        if event[0] == 'ai'
    ]
    tool_events = [
        event
        for event in events
        if event[0] == 'tool_call'
    ]
    ai_message: Optional[Message] = {
        'role': 'ai',
        'text': ''.join([e[2] for e in ai_events if e[1] == 'text_event']),
        'cost': cost,
        'raw': raw,
    } if len(ai_events) > 0 else None
    tool_message: Optional[MessageToolCall] = {
        'role': 'tool_call',
        'tools': [e[2] for e in tool_events if e[1] == 'tool_call_event'],
        'cost': cost,
        'raw': raw,
    } if len(tool_events) > 0 else None
    if ai_message and tool_message:
        ai_message['cost'] = None  # <-- we don't want to double-count this
        ai_message['raw'] = None
        return [ai_message, tool_message]
    elif ai_message:
        return [ai_message]
    elif tool_message:
        return [tool_message]
    else:
        raise ValueError('no events')


def _log_dumps(val: Any) -> str:
    if isinstance(val, dict):
        return json.dumps(val)
    else:
        return str(val)


class LasagnaNVIDIA(Model):
    def __init__(self, model: str, **model_kwargs: Dict[str, Any]):
        known_model_names = [m['formal_name'] for m in NVIDIA_KNOWN_MODELS]
        if model not in known_model_names:
            _LOG.warning(f'untested model: {model} (may or may not work)')
        self.model = model
        self.model_kwargs = copy.deepcopy(model_kwargs or {})
        self.n_retries: int = cast(int, self.model_kwargs['retries']) if 'retries' in self.model_kwargs else 3
        if not isinstance(self.n_retries, int) or self.n_retries < 0:
            raise ValueError(f"model_kwargs['retries'] must be a non-negative integer (got {self.model_kwargs['retries']})")

    def config_hash(self) -> str:
        return recursive_hash(None, {
            'provider': 'nvidia',
            'model': self.model,
            'model_kwargs': self.model_kwargs,
        })

    def _make_client(self) -> AsyncOpenAI:
        default_api_key = os.environ.get('NGC_API_KEY', None)
        default_base_url = 'https://integrate.api.nvidia.com/v1'
        api_key: Union[str, None] = cast(str, self.model_kwargs['api_key']) if 'api_key' in self.model_kwargs else default_api_key
        base_url: Union[str, None] = cast(str, self.model_kwargs['base_url']) if 'base_url' in self.model_kwargs else default_base_url
        client = AsyncOpenAI(
            api_key  = api_key,
            base_url = base_url,
        )
        return client

    async def _run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool,
    ) -> List[Message]:
        tool_choice: Union[ChatCompletionToolChoiceOptionParam, NotGiven]
        if force_tool:
            if len(tools) == 0:
                raise ValueError(f"When `force_tool` is set, you must pass at least one tool!")
            elif len(tools) == 1:
                tool_choice = {
                    "type": "function",
                    "function": {"name": tools[0].__name__},
                }
            else:
                tool_choice = 'required'  # <-- model must use a tool, but is allowed to choose which one on its own
        else:
            tool_choice = NOT_GIVEN  # <-- if tools given, the model can choose to use them or not

        tools_spec = _convert_to_openai_tools(tools)

        openai_messages = await _convert_to_openai_messages(messages)

        logprobs: Union[bool, NotGiven] = cast(bool, self.model_kwargs['logprobs']) if 'logprobs' in self.model_kwargs else NOT_GIVEN
        top_logprobs: Union[int, NotGiven] = cast(int, self.model_kwargs['top_logprobs']) if 'top_logprobs' in self.model_kwargs else (5 if logprobs is True else NOT_GIVEN)
        frequency_penalty: Union[float, NotGiven] = cast(float, self.model_kwargs['frequency_penalty']) if 'frequency_penalty' in self.model_kwargs else NOT_GIVEN
        presence_penalty: Union[float, NotGiven] = cast(float, self.model_kwargs['presence_penalty']) if 'presence_penalty' in self.model_kwargs else NOT_GIVEN
        max_tokens: Union[int, NotGiven] = cast(int, self.model_kwargs['max_tokens']) if 'max_tokens' in self.model_kwargs else 1024
        stop: Union[List[str], NotGiven] = cast(List[str], self.model_kwargs['stop']) if 'stop' in self.model_kwargs else NOT_GIVEN
        temperature: Union[float, NotGiven] = cast(float, self.model_kwargs['temperature']) if 'temperature' in self.model_kwargs else NOT_GIVEN
        top_p: Union[float, NotGiven] = cast(float, self.model_kwargs['top_p']) if 'top_p' in self.model_kwargs else NOT_GIVEN
        user: Union[str, NotGiven] = cast(str, self.model_kwargs['user']) if 'user' in self.model_kwargs else NOT_GIVEN

        _LOG.info(f"Invoking {self.model} with:\n  messages: {_log_dumps(openai_messages)}\n  tools: {_log_dumps(tools_spec)}\n  tool_choice: {tool_choice}")

        client = self._make_client()
        completion: AsyncIterator[ChatCompletionChunk] = await client.chat.completions.create(
            model        = self.model,
            messages     = openai_messages,
            tools        = tools_spec,
            tool_choice  = tool_choice,
            stream       = True,
            #stream_options = {'include_usage': True},
            logprobs     = logprobs,
            top_logprobs = top_logprobs,
            frequency_penalty = frequency_penalty,
            presence_penalty = presence_penalty,
            max_tokens = max_tokens,
            stop = stop,
            temperature = temperature,
            top_p = top_p,
            user = user,
        )

        raw_stream, rt_stream = adup(completion)

        events: List[EventPayload] = []

        async for event in _process_output_stream(_extract_deltas(rt_stream)):
            events.append(event)
            await event_callback(event)

        raw_payload = [v async for v in raw_stream]

        new_messages = _build_messages_from_openai_payload(raw_payload, events)

        _LOG.info(f"Finished {self.model} with usage: {_log_dumps(new_messages[-1].get('cost'))}")

        return new_messages

    async def _retrying_run_once(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool,
    ) -> List[Message]:
        last_error: Union[APIError, None] = None
        assert self.n_retries + 1 > 0   # <-- we know this is true from the check in __init__
        for delay_on_error in exponential_backoff_retry_delays(self.n_retries + 1):
            try:
                return await self._run_once(
                    event_callback = event_callback,
                    messages = messages,
                    tools = tools,
                    force_tool = force_tool,
                )
            except APIError as e:
                # Some errors should be retried, some should not. Below
                # is the logic to decide when to retry vs when to not.
                # It's likely this will change as we get more usage and see
                # where NVIDIA tends to fail, and when we know more what is
                # recoverable vs not.
                last_error = e
                if e.type == 'invalid_request_error':
                    if e.code == 'context_length_exceeded':
                        raise
                    elif e.code == 'invalid_api_key':
                        raise
                    else:
                        raise
                elif e.type == 'server_error':
                    pass  # <-- we will retry this one! I've seen these work when you just try again.
                else:
                    pass  # <-- this must be one we don't know about yet, so ... recoverable, maybe?
                if delay_on_error > 0.0:
                    _LOG.warning(f"Got a maybe-recoverable error (will retry in {delay_on_error:.2f} seconds): {e}")
                    await asyncio.sleep(delay_on_error)
        assert last_error is not None   # <-- we know this is true because `n_retries + 1 > 0`
        raise last_error

    async def run(
        self,
        event_callback: EventCallback,
        messages: List[Message],
        tools: List[Callable],
        force_tool: bool = False,
        max_tool_iters: int = 5,
    ) -> List[Message]:
        messages = [*messages]  # shallow copy
        new_messages: List[Message] = []
        for _ in range(max_tool_iters):
            new_messages_here = await self._retrying_run_once(
                event_callback = event_callback,
                messages       = messages,
                tools          = tools,
                force_tool     = force_tool,
            )
            tools_map = {tool.__name__: tool for tool in tools}
            new_messages.extend(new_messages_here)
            messages.extend(new_messages_here)
            tools_results = await handle_tools(new_messages_here, tools_map)
            if tools_results is None:
                break
            for tool_result in tools_results:
                await event_callback(('tool_res', 'tool_res_event', tool_result))
            tool_response_message = build_tool_response_message(tools_results)
            new_messages.append(tool_response_message)
            messages.append(tool_response_message)
        return new_messages



import pytest

from lasagna.types import (
    Message,
    ToolCall,
    EventPayload,
)

from lasagna.stream import fake_async

from lasagna.lasagna_openai import (
    _extract_deltas,
    _process_output_stream,
    _convert_to_openai_tools,
    _convert_to_openai_messages,
    _build_messages_from_openai_payload,
)

from typing import List

from openai.types.chat import (
    ChatCompletionChunk,
)


SAMPLE_TEXT_STREAM: List[ChatCompletionChunk] = [
    ChatCompletionChunk.model_validate(v)
    for v in [
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "",
                        "role": "assistant"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "Hello"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " Ryan"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "!"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " "
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "5"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " multiplied"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " by"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " "
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "7"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " is"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": " "
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "35"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {
                        "content": "."
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaHQcqZ177ddacDIf2nrdzbjoJRH",
            "choices": [
                {
                    "delta": {},
                    "finish_reason": "stop",
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713739736,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            'id': 'chatcmpl-9ViswdADZmieYlPfvYf7J2UgTgFC7',
            'choices': [],
            'created': 1717347734,
            'model': 'gpt-3.5-turbo-0125',
            'object': 'chat.completion.chunk',
            'system_fingerprint': None,
            'usage': {'completion_tokens': 22, 'prompt_tokens': 21, 'total_tokens': 43},
        },
    ]
]


SAMPLE_TOOL_STREAM: List[ChatCompletionChunk] = [
    ChatCompletionChunk.model_validate(v)
    for v in [
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "content": None,
                        "role": "assistant"
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 0,
                                "id": "call_x7zmzwKI0LrwDF2xVMcfzXzN",
                                "function": {
                                    "arguments": "",
                                    "name": "multiply"
                                },
                                "type": "function"
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 0,
                                "function": {
                                    "arguments": "{\"a\""
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 0,
                                "function": {
                                    "arguments": ": 5, "
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 0,
                                "function": {
                                    "arguments": "\"b\": 7"
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 0,
                                "function": {
                                    "arguments": "}"
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 1,
                                "id": "call_33vMBGeVd96A9BhW6H3r8jHb",
                                "function": {
                                    "arguments": "",
                                    "name": "multiply"
                                },
                                "type": "function"
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 1,
                                "function": {
                                    "arguments": "{\"a\""
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 1,
                                "function": {
                                    "arguments": ": 8, "
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 1,
                                "function": {
                                    "arguments": "\"b\": 1"
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {
                        "tool_calls": [
                            {
                                "index": 1,
                                "function": {
                                    "arguments": "01}"
                                }
                            }
                        ]
                    },
                    "finish_reason": None,
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        },
        {
            "id": "chatcmpl-9GaOnH7T5O8VitWZryKt1qWy77GW8",
            "choices": [
                {
                    "delta": {},
                    "finish_reason": "tool_calls",
                    "index": 0,
                    "logprobs": None
                }
            ],
            "created": 1713740193,
            "model": "gpt-3.5-turbo-0125",
            "object": "chat.completion.chunk",
            "system_fingerprint": "fp_c2295e73ad"
        }
    ]
]


CORRECT_PARSED_TOOLS: List[ToolCall] = [
    {
        "call_id": "call_x7zmzwKI0LrwDF2xVMcfzXzN",
        "function": {
            "arguments": "{\"a\": 5, \"b\": 7}",
            "name": "multiply"
        },
        "call_type": "function"
    },
    {
        "call_id": "call_33vMBGeVd96A9BhW6H3r8jHb",
        "function": {
            "arguments": "{\"a\": 8, \"b\": 101}",
            "name": "multiply"
        },
        "call_type": "function"
    }
]


@pytest.mark.asyncio
async def test_extract_deltas():
    stream = fake_async(SAMPLE_TEXT_STREAM[:2] + SAMPLE_TEXT_STREAM[-2:])
    vals = [(d.to_dict(), s) async for d, s in _extract_deltas(stream)]
    assert vals == [
        ({'content': '', 'role': 'assistant'}, None),
        ({'content': 'Hello'}, None),
        ({}, 'stop'),
    ]

    stream = fake_async(SAMPLE_TOOL_STREAM[:2] + SAMPLE_TOOL_STREAM[-1:])
    vals = [(d.to_dict(), s) async for d, s in _extract_deltas(stream)]
    assert vals == [
        ({'content': None, 'role': 'assistant'}, None),
        ({'tool_calls': [{
            'index': 0,
            'id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN',
            'function': {'arguments': '', 'name': 'multiply'},
            'type': 'function',
        }]}, None),
        ({}, 'tool_calls'),
    ]


@pytest.mark.asyncio
async def test_process_output_stream__text():
    stream = _process_output_stream(_extract_deltas(fake_async(SAMPLE_TEXT_STREAM)))
    texts: List[str] = []
    async for role, type_, text in stream:
        assert role == 'ai'
        assert type_ == 'text_event'
        assert isinstance(text, str)
        texts.append(text)
    text = ''.join(texts)
    assert text == 'Hello Ryan! 5 multiplied by 7 is 35.'


@pytest.mark.asyncio
async def test_process_output_stream__tool():
    stream = _process_output_stream(_extract_deltas(fake_async(SAMPLE_TOOL_STREAM)))
    texts: List[str] = []
    tool_calls: List[ToolCall] = []
    async for event in stream:
        assert event[0] == 'tool_call'
        if event[1] == 'text_event':
            texts.append(event[2])
        elif event[1] == 'tool_call_event':
            tool_calls.append(event[2])
        else:
            assert False, event[1]
    text = ''.join(texts)
    assert text == 'multiply({"a": 5, "b": 7})\nmultiply({"a": 8, "b": 101})'
    assert tool_calls == CORRECT_PARSED_TOOLS

    stream = _process_output_stream(_extract_deltas(fake_async(SAMPLE_TOOL_STREAM[:6])))
    texts: List[str] = []
    tool_calls: List[ToolCall] = []
    async for event in stream:
        assert event[0] == 'tool_call'
        if event[1] == 'text_event':
            texts.append(event[2])
        elif event[1] == 'tool_call_event':
            tool_calls.append(event[2])
        else:
            assert False, event[1]
    text = ''.join(texts)
    assert text == 'multiply({"a": 5, "b": 7})'
    assert tool_calls == CORRECT_PARSED_TOOLS[:1]


@pytest.mark.asyncio
async def test_process_output_stream__text_and_tool():
    # The model can start with text and switch to tools!
    stream = _process_output_stream(_extract_deltas(fake_async(
        SAMPLE_TEXT_STREAM[:-2] + SAMPLE_TOOL_STREAM[1:]
    )))
    texts: List[str] = []
    tool_calls: List[ToolCall] = []
    async for event in stream:
        if event[0] == 'ai':
            assert event[1] == 'text_event'
            texts.append(event[2])
        elif event[0] == 'tool_call':
            if event[1] == 'text_event':
                texts.append(event[2])
            elif event[1] == 'tool_call_event':
                tool_calls.append(event[2])
            else:
                assert False, event[1]
        else:
            assert False, event[0]
    text = ''.join(texts)
    assert text == 'Hello Ryan! 5 multiplied by 7 is 35.\n\nmultiply({"a": 5, "b": 7})\nmultiply({"a": 8, "b": 101})'
    assert tool_calls == CORRECT_PARSED_TOOLS


def test_convert_to_openai_tools():
    x = 5
    def mytool(a, b):
        """
        Does a thing.
        :param: a: str: what to do
        :param: b: int: when to do it
        """
        return a, b, x
    def my_other_tool(how):
        """Something else
        that does stuff.

        :param: how: enum good bad: whether to do the thing in a good or bad way
        """
        return how
    def thing_with_optional_params(first, second=True, third=1):
        """This tool has optional params.
        :param: first: float: a number of sorts
        :param: second: bool: (optional) pass if you want
        :param: third: int: (optional) something else that's optional
        """
        return first, second, third
    spec = _convert_to_openai_tools([mytool, my_other_tool, thing_with_optional_params])
    assert spec
    assert len(spec) == 3
    assert spec[0] == {
        'type': 'function',
        'function': {
            'name': 'mytool',
            'description': 'Does a thing.',
            'parameters': {
                'type': 'object',
                'properties': {
                    'a': {
                        'type': 'string',
                        'description': 'what to do',
                    },
                    'b': {
                        'type': 'integer',
                        'description': 'when to do it',
                    },
                },
                'required': ['a', 'b'],
            },
        },
    }
    assert spec[1] == {
        'type': 'function',
        'function': {
            'name': 'my_other_tool',
            'description': 'Something else that does stuff.',
            'parameters': {
                'type': 'object',
                'properties': {
                    'how': {
                        'type': 'string',
                        'enum': ['good', 'bad'],
                        'description': 'whether to do the thing in a good or bad way',
                    },
                },
                'required': ['how'],
            },
        },
    }
    assert spec[2] == {
        'type': 'function',
        'function': {
            'name': 'thing_with_optional_params',
            'description': 'This tool has optional params.',
            'parameters': {
                'type': 'object',
                'properties': {
                    'first': {
                        'type': 'number',
                        'description': 'a number of sorts',
                    },
                    'second': {
                        'type': 'boolean',
                        'description': '(optional) pass if you want',
                    },
                    'third': {
                        'type': 'integer',
                        'description': '(optional) something else that\'s optional',
                    },
                },
                'required': ['first'],
            },
        },
    }


@pytest.mark.asyncio
async def test_convert_to_openai_messages():
    messages: List[Message] = [
        {
            'role': 'INVALID', # type: ignore
            'text': 'bla',
            'cost': None,
            'raw': None,
        },
    ]
    with pytest.raises(ValueError):
        await _convert_to_openai_messages(messages)

    messages: List[Message] = [
        {'role': 'system', 'text': 'be nice', 'cost': None, 'raw': None},
        {'role': 'human', 'text': 'hi', 'cost': None, 'raw': None},
        {'role': 'ai', 'text': 'oh hi', 'cost': None, 'raw': None},
        {'role': 'human', 'text': 'here is a picture', 'media': [{'type': 'image', 'image': 'http://example.com/img.png'}], 'cost': None, 'raw': None},
        {'role': 'ai', 'text': 'thanks!', 'cost': None, 'raw': None},
        {'role': 'human', 'text': 'here are two', 'media': [{'type': 'image', 'image': 'http://example.com/img.png'}, {'type': 'image', 'image': 'http://example.com/img2.png'}], 'cost': None, 'raw': None},
        {'role': 'ai', 'text': 'double thanks!', 'cost': None, 'raw': None},
    ]
    ms = await _convert_to_openai_messages(messages)
    assert ms == [
        {'role': 'system', 'content': 'be nice'},
        {'role': 'user', 'content': [{'type': 'text', 'text': 'hi'}]},
        {'role': 'assistant', 'content': 'oh hi'},
        {'role': 'user', 'content': [{'type': 'text', 'text': 'here is a picture'}, {'type': 'image_url', 'image_url': {'url': 'http://example.com/img.png'}}]},
        {'role': 'assistant', 'content': 'thanks!'},
        {'role': 'user', 'content': [{'type': 'text', 'text': 'here are two'}, {'type': 'image_url', 'image_url': {'url': 'http://example.com/img.png'}}, {'type': 'image_url', 'image_url': {'url': 'http://example.com/img2.png'}}]},
        {'role': 'assistant', 'content': 'double thanks!'},
    ]

    messages: List[Message] = [{
        'role': 'tool_call',
        'tools': [
            {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
            {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
        ],
        'cost': None,
        'raw': None,
    }]
    ms = await _convert_to_openai_messages(messages)
    assert ms == [{
        'role': 'assistant',
        'content': None,
        'tool_calls': [
            {'id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'type': 'function'},
            {'id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'type': 'function'},
        ],
    }]

    messages: List[Message] = [{
        'role': 'tool_res',
        'tools': [
            {'call_id': '1002', 'result': 10.8 },
            {'call_id': '1003', 'result': "hihi" },
        ],
        'cost': None,
        'raw': None,
    }]
    ms = await _convert_to_openai_messages(messages)
    assert ms == [
        {
            'role': 'tool',
            'content': "10.8",
            'tool_call_id': "1002",
        },
        {
            'role': 'tool',
            'content': "hihi",
            'tool_call_id': "1003",
        },
    ]

    messages: List[Message] = [
        {
            'role': 'ai',
            'text': "I'll use my tools!",
            'cost': None,
            'raw': None,
        },
        {
            'role': 'tool_call',
            'tools': [
                {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
                {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
            ],
            'cost': None,
            'raw': None,
        },
    ]
    ms = await _convert_to_openai_messages(messages)
    assert ms == [{
        'role': 'assistant',
        'content': "I'll use my tools!",
        'tool_calls': [
            {'id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'type': 'function'},
            {'id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'type': 'function'},
        ],
    }]


def test_build_messages_from_openai_payload():
    events: List[EventPayload] = [
        ('ai', 'text_event', 'Hello'),
        ('ai', 'text_event', ' Ryan'),
    ]
    messages = _build_messages_from_openai_payload([], events)
    assert messages == [{
        'role': 'ai',
        'text': 'Hello Ryan',
        'cost': None,
        'raw': [],
    }]

    tool_calls: List[ToolCall] = [
        {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
        {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
    ]
    events: List[EventPayload] = [
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"a"'),
        ('tool_call', 'text_event', ': 5, '),
        ('tool_call', 'text_event', '"b": 7'),
        ('tool_call', 'text_event', '}'),
        ('tool_call', 'text_event', ')\n'),
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"a"'),
        ('tool_call', 'text_event', ': 8, '),
        ('tool_call', 'text_event', '"b": 1'),
        ('tool_call', 'text_event', '01}'),
        ('tool_call', 'text_event', ')'),
        ('tool_call', 'tool_call_event', tool_calls[0]),
        ('tool_call', 'tool_call_event', tool_calls[1]),
    ]
    messages = _build_messages_from_openai_payload([], events)
    assert messages == [{
        'role': 'tool_call',
        'tools': [
            {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
            {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
        ],
        'cost': None,
        'raw': [],
    }]

    tool_calls: List[ToolCall] = [
        {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
        {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
    ]
    events: List[EventPayload] = [
        ('ai', 'text_event', 'Hello'),
        ('ai', 'text_event', ' Ryan'),
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"a"'),
        ('tool_call', 'text_event', ': 5, '),
        ('tool_call', 'text_event', '"b": 7'),
        ('tool_call', 'text_event', '}'),
        ('tool_call', 'text_event', ')\n'),
        ('tool_call', 'text_event', 'multiply('),
        ('tool_call', 'text_event', '{"a"'),
        ('tool_call', 'text_event', ': 8, '),
        ('tool_call', 'text_event', '"b": 1'),
        ('tool_call', 'text_event', '01}'),
        ('tool_call', 'text_event', ')'),
        ('tool_call', 'tool_call_event', tool_calls[0]),
        ('tool_call', 'tool_call_event', tool_calls[1]),
    ]
    messages = _build_messages_from_openai_payload(SAMPLE_TEXT_STREAM[-3:], events)
    assert messages == [
        {
            'role': 'ai',
            'text': 'Hello Ryan',
            'cost': None,
            'raw': None,
        },
        {
            'role': 'tool_call',
            'tools': [
                {'call_id': 'call_x7zmzwKI0LrwDF2xVMcfzXzN', 'function': {'arguments': '{"a": 5, "b": 7}', 'name': 'multiply'}, 'call_type': 'function'},
                {'call_id': 'call_33vMBGeVd96A9BhW6H3r8jHb', 'function': {'arguments': '{"a": 8, "b": 101}', 'name': 'multiply'}, 'call_type': 'function'},
            ],
            'cost': {
                'input_tokens': 21,
                'output_tokens': 22,
                'total_tokens': 43,
            },
            'raw': [
                v.to_dict()
                for v in SAMPLE_TEXT_STREAM[-3:]
            ],
        },
    ]
